#include <stdio.h>
#include <stdarg.h>
#define TAME_INTERNAL
#include "tame.h"

// Runtime

static List* blocked;    // List(Rendezvous*)

void
tame_init()
{
	printf("Tame init!\n");
}

void
tame_loop()
{
	List *l, *ll;
	Rendezvous *r;

	printf("Tame loop!\n");
	while(blocked){
		ll = blocked;
		blocked = nil;
		for(l=ll; l; l=l->next){
			r = l->x;
			if(!r->ready || (r->waiting && r->flag == TameResumeAll)){
				blocked = listappend(nil, blocked, r);
				continue;
			}
			r->cl->__cb(r->cl);
		}
	}
	printf("Tame done!\n");
}

// Interface

Event*
tame_mkevent(Rendezvous *r, int id, int nvalues, ...)
{
	Event *e;
	int i;
	va_list va;

	assert(r);
	e = emalloc(sizeof(*e) + nvalues*sizeof(void*));
	e->r = r;
	e->id = id;
	e->nvalues = nvalues;
	va_start(va, nvalues);
	for(i=0; i < nvalues; i++)
		e->values[i] = va_arg(va, void*);
	va_end(va);
	e->r->waiting = listinsert(nil, e->r->waiting, e);
	return e;
}

void
tame_unblock(Event *e)
{
	List *l, *ll;
	Rendezvous *r;

	assert(e);
	assert(e->r);
	r = e->r;
	ll = ZList;
	for(l=e->r->waiting; l; l=l->next){
		if(l->x == e){
			if(ll)
				ll->next = l->next;
			else
				ll = l->next;
			r->ready = listappend(nil, r->ready, e);
			break;
		}
		ll = listappend(nil, ll, l->x);
	}
	r->waiting = ll;
}

int
tame_nleft(Rendezvous *r)
{
	assert(r);
	return listlen(r->waiting) + listlen(r->ready);
}

bool
tame_allready(Rendezvous *r)
{
	assert(r);
	return r->waiting == ZList;
}

bool
tame_getready(Rendezvous *r, int *id)
{
	Event *e;

	assert(r);
	if(!r->ready)
		return false;
	e = listpeek(r->ready);
	r->ready = listpop(r->ready);
	if(id)
		*id = e->id;
	return true;
}

void
tame_registercb(Closure *cl, Rendezvous *r, TameResumeFlag f)
{
	assert(r);
	assert(cl);
	r->cl = cl;
	r->flag = f;
	blocked = listappend(nil, blocked, r);
}

void
tame_trigger__void(Event *e)
{
	assert(e);
	// non-void triggers do (*e->values[n] = *vn) here
	tame_unblock(e);
}

