#include <assert.h>
#include <string.h>
#include <emalloc.h>
#include <list2.h>

typedef _Bool bool;
static const bool true=1;
static const bool false=0;

#ifndef nil
#define nil (void*)0
#endif

// TODO: refcounting?
typedef struct Closure Closure;
typedef struct Event Event;
typedef struct Rendezvous Rendezvous;
typedef enum TameResumeFlag TameResumeFlag;
typedef void (*tame_cb)(void *cl);

enum TameResumeFlag
{
	TameResumeAny,
	TameResumeAll,
};

struct Closure
{
	tame_cb __cb;
	int     __resume;
	// the closure follows
	// (this is auto-generated for each tamed event function)
};

struct Event
{
	Rendezvous *r;
	int id;
	int nvalues;
	void* values[];   // trigger values
};
struct Rendezvous
{
	Closure *cl;
	TameResumeFlag flag;
	List *waiting;    // List(Event*), not-yet triggered
	List *ready;      // List(Event*), already triggered, not-yet witnessed
};

// The exposed interface is:
//   event<>*      mkevent(Rendezvous* r, int id)
//   event<T1,..>* mkevent(Rendezvous* r, int id, T1 *v1, T2 *v2, ...)
//   void          trigger(event<>* e)
//   void          trigger(event<>* e, T1 v1, T2 v2, ...)

void   tame_init(void);
void   tame_loop(void);

// Bogus
#ifndef TAME_INTERNAL
Event* mkevent(Rendezvous *r, int id, ...);
void   trigger(Event *e, ...);
#endif

// This is the internal interface:
Event* tame_mkevent(Rendezvous *r, int id, int nvalues, ...);
void   tame_unblock(Event *e);
int    tame_nleft(Rendezvous *r);
bool   tame_allready(Rendezvous *r);
bool   tame_getready(Rendezvous *r, int *id);
void   tame_registercb(Closure *cl, Rendezvous *r, TameResumeFlag f);
void   tame_trigger__void(Event *e);

// Autogenerated (on demand):
//   tame_trigger__1(Event* e, int v1)
//   tame_trigger__2(Event* e, float v1)
//   tame_trigger__3(Event* e, float v1, int v2)
//   ...

// Runtime has:
//   set pair (waiting closure, waited-on rendezvous)
//   when an event triggers, it notifies a rendezvous which wakes a closure


