// Builtin functions (escapes to native C).
// Somewhat haphazard at the moment.

#include "grammar.h"
extern "C" {
#include "regexp9.h"
}
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <stdio.h>

#define listappend(x, y) listappend((Arena*)nil, x, y)
#define list1(x) list1((Arena*)nil, x)
#define list2(x, y) list2((Arena*)nil, x, y)
#define list3(x, y, z) list3((Arena*)nil, x, y, z)
#define list4(x, y, z, w) list4((Arena*)nil, x, y, z, w)

void builtinexit(void *v);

static int 
readn(int fd, char *buf, int n)
{
	int tot, m;
	
	for(tot=0; tot<n; tot+=m){
		m = read(fd, buf+tot, n-tot);
		if(m <= 0)
			break;
	}
	return tot;
}

void
builtinreadfile(void *v)
{
	struct fnframe {
		char *file;
		Name *data;
	} *f = static_cast<fnframe*>(v);
	int fd;
	struct stat st;
	char *buf;
	
	if((fd = open(f->file, O_RDONLY)) < 0){
		*f->data = nameof("");
		return;
	}
	if(fstat(fd, &st) < 0){
		close(fd);
		*f->data = nameof("");
		return;
	}
	buf = static_cast<char*>(emalloc(st.st_size+1));
	if(readn(fd, buf, st.st_size) != st.st_size){
		close(fd);
		free(buf);
		*f->data = nameof("");
		return;
	}
	close(fd);
	*f->data = nameof(buf);
	free(buf);
}

void
builtinwritefile(void *v)
{
	struct fnframe {
		char *data;
		char *file;
		int *ret;
	} *f = static_cast<fnframe*>(v);
	int fd, n, m;
	char *p;

	if(strcmp(f->file, "/dev/stdout") == 0){
		fd = 1;
	}else if(strcmp(f->file, "/dev/stderr") == 0){
		fd = 2;
	}else if((fd = creat(f->file, 0666)) < 0){
		*f->ret = -1;
		return;
	}
	p = f->data;
	if(p == nil)
		p = "";
	n = strlen(p);
	while(n > 0){
		m = write(fd, p, n);
		if(m <= 0)
			break;
		p += m;
		n -= m;
	}
	if(fd > 2)
		close(fd);
	if(n > 0)
		*f->ret = -1;
	else
		*f->ret = 0;
}

typedef struct StringList StringList;
struct StringList
{
	StringList *next;
	char *s;
};

void
builtinreadpipe(void *v)
{
	char **argv;
	StringList *l;
	int i, fd, n, m, tot, p[2], status;
	pid_t pid;
	char *buf;

	struct R {
		Name str;
		int code;
	};
	struct fnframe {
		StringList *cmd;
		struct R *data;
	} *f = static_cast<fnframe*>(v);
	
	f->data->str = nameof("");
	f->data->code = -1;
	i = 0;
	for(l=f->cmd; l; l=l->next)
		i++;
	if(i == 0){
		fprint(2, "readpipe: no command\n");
		return;
	}
	argv = static_cast<char**>(emalloc((i+1) * sizeof(argv[0])));
	i = 0;
	for(l=f->cmd; l; l=l->next)
		argv[i++] = l->s;
	argv[i] = nil;
	
	if(pipe(p) < 0)
		panic("pipe");
	pid = fork();
	if(pid == 0){
		close(p[0]);
		dup2(p[1], 1);
		execvp(argv[0], argv);
		panic("exec cpp: %r");
		_exit(1);
	}
	close(p[1]);
	fd = p[0];

	tot = 0;
	m = 2048;
	buf = static_cast<char*>(emalloc(m));
	for(;;){
		m *= 2;
		buf = static_cast<char*>(erealloc(buf, m));
		n = readn(fd, buf + tot, m - tot);
		if(n <= 0)
			break;
		tot += n;
		if(tot < m)
			break;
	}
	buf[tot] = 0;
	f->data->str = nameof(buf);
	free(buf);
	close(fd);

	waitpid(pid, &status, 0);
	if(WIFEXITED(status))
		f->data->code = WEXITSTATUS(status);
}


void
builtingets(void *v)
{
	struct fnframe {
		char **out;
	} *f = static_cast<fnframe*>(v);
	char buf[1024];
	char *p;
	
	p = fgets(buf, sizeof buf, stdin);
	if(p == nil){
		*f->out = "";
		return;
	}
	p[strlen(p)-1] = 0;
	*f->out = static_cast<char*>(ecopy(p, strlen(p)+1));
}

void
builtinabort(void *v)
{
	abort();
}

void
builtinastsplit(void *v)
{
	struct fnframe {
		Ast *ast;
		Array **out;
	} *f = static_cast<fnframe*>(v);

	*f->out = astsplit(f->ast);
}

void
builtinastlastkid(void *v)
{
	struct fnframe {
		Ast *ast;
		Ast **out;
	} *f = static_cast<fnframe*>(v);

	*f->out = astlastkid(f->ast);
}

void
builtinastjoin(void *v)
{
	struct fnframe {
		Array *array;
		Ast *ast;
		Ast **out;
	} *f = static_cast<fnframe*>(v);
	if(f->array == nil)
		*f->out = f->ast;
	else
		*f->out = astjoin(f->ast, f->array->n, static_cast<Ast**>(f->array->v));
}

void
builtinastcopy(void *v)
{
	struct fnframe {
		Ast *ast;
		Ast **out;
	} *f = static_cast<fnframe*>(v);

	*f->out = astcopy(f->ast, 0);
}

void
builtinastdeepcopy(void *v)
{
	struct fnframe {
		Ast *ast;
		Ast **out;
	} *f = static_cast<fnframe*>(v);

	*f->out = astdeepcopy(f->ast);
}

void
builtinastremovecopiedfrom(void *v)
{
	struct fnframe {
		int deep;
		Ast *ast;
	} *f = static_cast<fnframe*>(v);

	if(f->deep)
		astremovecopiedfrom(f->ast);
	if(f->ast)
		f->ast->copiedfrom = nil;
}

void
builtinastlisp(void *v)
{
	struct fnframe {
		Ast *ast;
		Name *out;
	} *f = static_cast<fnframe*>(v);

	*f->out = nameprint("%#lA", f->ast);
}

void
builtinasttype(void *v)
{
	struct fnframe {
		Ast *ast;
		Name *out;
	} *f = static_cast<fnframe*>(v);

	if(f->ast == nil)
		*f->out = nameof("nil");
	else
		*f->out = nameprint("%T", f->ast->_specifictype);
}

void
builtinastisshared(void *v)
{
	struct fnframe {
		Ast *ast;
		int *out;
	} *f = static_cast<fnframe*>(v);

	// Note that 'op == AstString' --> 'isshared'
	*f->out = f->ast->isshared;
}

void
builtinfmt(void *v)
{
	struct fnframe {
		Name s;
		Fmt *fmt;
		void *out;
	} *f = static_cast<fnframe*>(v);

	if(f->s == nil || f->fmt == nil)
		return;
	if(f->fmt->flush == nil)
		fmtstrinit(f->fmt);
	fmtstrcpy(f->fmt, f->s);
}

void
builtinfmtflush(void *v)
{
	struct fnframe {
		Fmt *fmt;
		char **out;  // not Name*: premature optimization
	} *f = static_cast<fnframe*>(v);
	
	if(f->fmt == nil || f->fmt->flush == nil){
		*f->out = nil;
		return;
	}
	*f->out = fmtstrflush(f->fmt);
	memset(f->fmt, 0, sizeof *f->fmt);
}

// Fork!  Child returns 0 immediately.  Parent waits for child,
// then returns 1.  Good for testing that something should
// panic without actually killing the script.
void
builtinfork(void *v)
{
	struct fnframe {
		int *ret;
	} *f = static_cast<fnframe*>(v);
	int pid;

	pid = fork();
	switch(pid){
	case -1:
		*f->ret = -1;
		return;
	case 0:
		*f->ret = 0;
		return;
	default:
		waitpid(pid, nil, 0);
		*f->ret = 1;
		return;
	}
}

void
builtinsleep(void *v)
{
	struct fnframe {
		double s;
		void *out;
	} *f = static_cast<fnframe*>(v);
	struct timeval tv;
	
	tv.tv_sec = static_cast<int>(f->s);
	tv.tv_usec = static_cast<int>(fmod(f->s, 1) * 1000000);
	select(0, NULL, NULL, NULL, &tv);
}

void
builtinloadextension(void *v)
{
	struct fnframe {
		Name name;
		void *out;
	} *f = static_cast<fnframe*>(v);

	loadextension(f->name);
}

void
builtingsub(void *v)
{
	struct fnframe {
		char *repl;
		char *pattern;
		char *in;
		Name *out;
	} *f = static_cast<fnframe*>(v);
	Fmt fmt;
	char *p, *lastmatch, *repl, *pattern;
	char buf[1024];
	Resub m[10];
	Reprog *prog;
	
	repl = f->repl;
	if(repl == nil)
		repl = "";
	pattern = f->pattern;
	if(pattern == nil)
		pattern = "";
	prog = regcomp(pattern);

	fmtstrinit(&fmt);
	p = f->in;
	if(p == nil)
		p = "";
	lastmatch = nil;
	for(;;){
		memset(m, 0, sizeof m);
		if(!regexec(prog, p, m, nelem(m)))
			break;
		if(m[0].s.sp == lastmatch){
			if(*p == 0)
				break;
			memset(m, 0, sizeof m);
			if(!regexec(prog, p+1, m, nelem(m)))
				break;
		}
		if(m[0].s.sp > p)
			fmtprint(&fmt, "%.*s", utfnlen(p, m[0].s.sp - p), p);
		buf[0] = 0;
		regsub(f->repl, buf, sizeof buf, m, nelem(m));
		fmtstrcpy(&fmt, buf);
		lastmatch = m[0].s.sp;
		p = m[0].e.ep;
	}
	fmtstrcpy(&fmt, p);

	p = fmtstrflush(&fmt);
	*f->out = nameof(p);
	free(p);
		
	free(prog);
}

void
builtinlistsort(void *v)
{
	struct fnframe {
		ListNode *in;
		ListNode **out;
	} *f = static_cast<fnframe*>(v);

	if(f->out){
		List<void*> *l = listnodetolist(nil, f->in);
		l = listsort(l, (int(*)(void*,void*))nil);
		*f->out = listtolistnode(l);
		listfree(l);
	}
}

void
builtinlisteq(void *v)
{
	struct fnframe {
		ListNode *in0;
		ListNode *in1;
		int *out;
	} *f = static_cast<fnframe*>(v);
	List<void*> *l0, *l1;

	if(f->out){
		l0 = listnodetolist(nil, f->in0);
		l1 = listnodetolist(nil, f->in1);
		*f->out = listeq(l0, l1);
		listfree(l0);
		listfree(l1);
	}
}

void
builtinlistmap(void *v)
{
	panic("listmap should be emitted as xlistmap!");
}

void
builtinsetparent(void *v)
{
	struct fnframe {
		Ast *parent;
		Ast *term;
		void *out;
	} *f = static_cast<fnframe*>(v);

	assert(f->term->_parent == NoParentYet && !f->term->isshared);
	f->term->_parent = f->parent;
}

void
builtinunsignedrsh(void *v)
{
	struct fnframe {
		unsigned int bits;
		unsigned int num;
		unsigned int *out;
	} *f = static_cast<fnframe*>(v);

	*f->out = f->num << f->bits;
}

void
builtincescapestring(void *v)
{
	struct fnframe {
		Name delim;
		Name in;
		Name *out;
	} *f = static_cast<fnframe*>(v);
	int q;

	if(!f->delim || !f->delim[0])
		q = '"';
	else if(f->delim[1])
		panic("bad delimeter for cescapestring");
	else
		q = f->delim[0];
	*f->out = nameof(cescapestring(f->in, -1, q));
}

void
builtincunescapestring(void *v)
{
	struct fnframe {
		char *in;
		char **out;
	} *f = static_cast<fnframe*>(v);

	*f->out = nameof(cunescapestring(f->in, -1));
}

void
builtinord(void *v)
{
	struct fnframe {
		char *in;
		int *out;
	} *f = static_cast<fnframe*>(v);

	if(strlen(f->in) != 1)
		panic("ord string length != 1");

	*f->out = f->in[0];
}

static double timeofboot;

void
builtintimeofboot(void *v)
{
	struct fnframe {
		double *out;
	} *f = (fnframe*)v;

	*f->out = timeofboot;
}

void
builtintime(void *v)
{
	struct fnframe {
		double *out;
	} *f = (fnframe*)v;

	struct timeval tv;

	*f->out = 0;
	if(gettimeofday(&tv, 0) < 0)
		return;

	*f->out = tv.tv_sec + tv.tv_usec*1.e-6;
}

extern int nmkast;
void
builtinastcount(void *v)
{
	struct fnframe {
		int *out;
	} *f = (fnframe*)v;
	
	*f->out = nmkast;
}

// Install builtins.
void
builtininit(int argc, char **argv)
{
	extern Type *_typeast;
	extern Type *_typeaststring;
	extern Type *_typeline;
	extern Type *_typeframe;
	extern Type *_typescope;
	extern Type *_typesym;
	Array *a;
	Type *t, *s, *host, *typefmt, *structast;
	Var *v;
	int i, off;

	struct timeval tv;
	if(gettimeofday(&tv, 0) < 0)
		panic("gettimeofday() failed");
	timeofboot = tv.tv_sec + tv.tv_usec*1.e-6;

	// args: array string
	a = static_cast<Array*>(emalloc(sizeof *a));
	a->n = argc;
	a->sizeof1 = sizeof(char*);
	a->v = ecopy(argv, argc*sizeof(argv[0]));
	argv = static_cast<char**>(a->v);
	for(i=0; i<argc; i++)
		argv[i] = nameof(argv[i]);
	installbuiltindata("args", &a, typearray(typestring()));

	// not builtins, but we know about them
	_typeframe = typeunresolved(globalscope, nameof("Frame"));
	_typescope = typeunresolved(globalscope, nameof("Scope"));
	_typesym   = typeunresolved(globalscope, nameof("Sym"));

	List<Type*> *lt;
	List<Var*> *lv;
#define addstructvar(name, type, _isconst) \
	t = type; \
	v = mkVar(nameof(name), t); \
	v->off = off; \
	v->isconst = _isconst; \
	v->host = host; \
	off += sizeoftype(t); \
	lt = listappend(lt, t); \
	lv = listappend(lv, v)

	structast = mkTypeStruct(nil, nameof("Ast"), nil);
	_typeast  = mkTypeSyntax(nameof("Ast"), nil, structast, nil);
	_typeline = mkTypeStruct(nil, nameof("Line"), nil);

	structast->tagshash = mkhash<Name,List<Type*>*>();  // FIXME: put this elsewhere
	structast->discriminated = 1;
	structast->extensible = 1;
	_typeast->tagshash = mkhash<Name,List<Type*>*>();

	// struct Line {
	//   file : string;
	//   line : int;
	// }
	lt = nil;
	lv = nil;
	off = offsetof(ZLine, file);
	host = _typeline;
	addstructvar("file", typestring(), 0);
	addstructvar("line", typeint(), 0);

	t = _typeline;
	t->types = lt;
	t->vars  = lv;
	t->ctorargs = lt;
	installbuiltintype("Line", _typeline);
	_typeline = t;

	// extensible discriminated struct Ast {
	//   gramrule : int;        // read-only
	//   parent   : ptr Ast;    // read-only
	//   symname  : string;     // read-only
	//   line     : Line;
	// }
	lt = nil;
	lv = nil;
	off = offsetof(Ast, rule);
	host = structast;
	addstructvar("gramrule", typeptr(typedontcare()), 1);
	addstructvar("parent", typeptr(_typeast), 1);
	addstructvar("copiedfrom", typeptr(_typeast), 1);
	addstructvar("isshared", typebool(), 1);
	addstructvar("symname", typestring(), 1);
	addstructvar("line", _typeline, 0);

	t = structast;
	t->types = lt;
	t->vars  = lv;
	t->ctorargs = lt;
	t->reserved = offsetof(Ast, extensions);
	t->extensible = 1;
	t->discriminated = 1;
	t->tagshash = mkhash<Name,List<Type*>*>();
	installbuiltintype("Ast", _typeast);

	// attributes
	addstructattr(nameof("prev"), typeptr(typeast()), typeast()->structrep, 1);
	addstructattr(nameof("next"), typeptr(typeast()), typeast()->structrep, 1);
	fixglobalattrs();

	// tags
	_typeaststring = mkTypeSyntax(nameof("#string"), _typeast, structast, nil);
	addstructtag(_typeast, _typeaststring);
	addbuiltingramtags(_typeast, nil, nil);

	// check builtins
	assert(sizeoftype(structast) == sizeof(Ast));
	assert(sizeoftype(_typeast)  == sizeof(Ast));
	assert(sizeoftype(_typeline) == sizeof(ZLine));

	// Fmt
	t = mkTypeStruct(nil, nameof("Fmt"), nil);
	t->reserved = sizeof(Fmt);
	installbuiltintype("Fmt", t);
	typefmt = t;

#undef addstructvar

#define B(name, type)  installbuiltin(#name, builtin ## name, type)

	Type *ptrast = typeptr(typeast());

	B(readfile,   typefn(list1(typestring()), typestring()));
	B(writefile,  typefn(list2(typestring(), typestring()), typeint()));
	B(gets,       typefn(nil, typestring()));
	B(readpipe,   typefn(list1(typelist(typestring())), typetuple(list2(typestring(), typeint()))));
	B(timeofboot, typefn(nil, typefloat()));
	B(time,       typefn(nil, typefloat()));
	B(astcount,   typefn(nil, typeint()));

	B(abort,    typefn(nil, typetuple(nil)));
	B(exit,     typefn(list1(typeint()), typetuple(nil)));
	B(sleep,    typefn(list1(typefloat()), typetuple(nil)));
	B(fmt,      typefn(list2(typeptr(typefmt), typestring()), typevoid()));
	B(fmtflush, typefn(list1(typeptr(typefmt)), typestring()));
	B(fork,     typefn(nil, typeint()));
	B(gsub,     typefn(list3(typestring(), typestring(), typestring()), typestring()));

	B(loadextension,   typefn(list1(typestring()), typevoid()));
	B(unsignedrsh,     typefn(list2(typeint(), typeint()), typeint()));
	B(cescapestring,   typefn(list2(typestring(), typestring()), typestring()));
	B(cunescapestring, typefn(list1(typestring()), typestring()));
	B(ord,             typefn(list1(typestring()), typeint()));
	
	B(astsplit,    typefn(list1(ptrast), typearray(ptrast)));
	B(astjoin,     typefn(list2(ptrast, typearray(ptrast)), ptrast));
	B(astlastkid , typefn(list1(ptrast), ptrast));
	B(astcopy,     typefn(list1(ptrast), ptrast));
	B(astdeepcopy, typefn(list1(ptrast), ptrast));
	B(astremovecopiedfrom, typefn(list2(ptrast, typebool()), typetuple(nil)));
	B(astlisp,     typefn(list1(ptrast), typestring()));
	B(asttype,     typefn(list1(ptrast), typestring()));
	B(astisshared, typefn(list1(ptrast), typebool()));
	B(setparent,   typefn(list2(ptrast, ptrast), typevoid()));

#define FORALL(x, t)      typeforall(list1(x), t)
#define FORALL2(x, y, t)  typeforall(list2(x, y), t)

	t = mkTypeVar(nameof("T"));
	B(listsort, FORALL(t, typefn(list1(typelist(t)), typelist(t))));
	t = mkTypeVar(nameof("T"));
	B(listeq,   FORALL(t, typefn(list2(typelist(t), typelist(t)), typebool())));
	s = mkTypeVar(nameof("S"));
	t = mkTypeVar(nameof("T"));
	B(listmap,  FORALL2(s,t, typefn(list2(typelist(s), typefn(list1(s), t)), typelist(t))));


#undef B
#undef FORALL
}

