// Table-driven disassembler.
// Requires that each opcode has only one associated encoding.

#include "a.h"
#include "code.h"

typedef struct Disa Disa;
struct Disa
{
	Code *code;
	char *fmt;
};

Disa nametab[] =
{
#include "codenames.c"
};

// Perhaps this table should be automatically generated
// using comments in code.c?  That way the descriptions
// would be right next to the decoding code.
Disa distab[] =
{
	{ xjmp, "jmp %r" }, 
	{ xcjmp, "jc %l, %r" }, 
	{ xncjmp, "jnc %l, %r" }, 
	{ xloadp, "loadp %l, %p" }, 
	{ xloadi, "loadi %l, %i" }, 
	{ xloadf, "loadf %l, %f" }, 
	{ xmov4, "mov %l, %l" }, 
	{ xmovn, "movn %i, %l, %l" }, 
	{ xnew, "new %l, %t" },
	{ xalloc, "alloc %l, %i" },
	{ xlinkfn, "linkfn %l, %l, %p" },
	{ xcall, "call %l, %l" }, 
	{ xretcall, "retcall %l, %l" }, 
	{ xstop, "stop" },
	{ xprint, "print [%t] %l" }, 
	{ xret, "ret" },
	{ xaddi, "addi %l, %l, %l" },
	{ xaddf, "addf %l, %l, %l" },
	{ xsubi, "subi %l, %l, %l" },
	{ xsubf, "subf %l, %l, %l" },
	{ xmuli, "muli %l, %l, %l" },
	{ xmulf, "mulf %l, %l, %l" },
	{ xdivi, "divi %l, %l, %l" },
	{ xdivf, "divf %l, %l, %l" },
	{ xmodi, "modi %l, %l, %l" },
	{ xmodf, "modf %l, %l, %l" },
	{ xlshi, "lshi %l, %l, %l" },
	{ xrshi, "rshi %l, %l, %l" },
	{ xori, "ori %l, %l, %l" },
	{ xandi, "andi %l, %l, %l" },
	{ xxori, "xori %l, %l, %l" },
	{ xlti, "lti %l, %l, %l" },
	{ xltf, "ltf %l, %l, %l" },
	{ xgti, "gti %l, %l, %l" },
	{ xgtf, "gtf %l, %l, %l" },
	{ xlei, "lei %l, %l, %l" },
	{ xlef, "lef %l, %l, %l" },
	{ xgei, "gei %l, %l, %l" },
	{ xgef, "gef %l, %l, %l" },
	{ xeqi, "eqi %l, %l, %l" },
	{ xeqf, "eqf %l, %l, %l" },
	{ xnei, "nei %l, %l, %l" },
	{ xnef, "nef %l, %l, %l" },
	{ xnegi, "negi %l, %l" },
	{ xnegf, "negf %l, %l" },
	{ xposi, "posi %l, %l" },
	{ xposf, "posf %l, %l" },
	{ xtwidi, "twidi %l, %l" },
	{ xnoti, "noti %l, %l" },
	{ xmkframe, "mkframe %l, %l" },
	{ xcall, "call %l, %l" },
	{ xlea, "lea %l, %l" },
	{ xleax, "leax %l, %l, %t" },
	{ xcopystruct, "copystruct %l, %l, %t, %t" },
	{ xdestructure, "destructure %l, %l, %t, %p, %i, %p, %p, %l, %p" },
	{ xrestructure, "restructure %l, %p, %i, %p, %p, %l" },
	{ xrestructslot, "restructslot %l, %t, %l" },
	{ xsamestructure, "samestructure %l, %l, %l" },
	{ xnzf, "nzf %l, %l" },
	{ xnzs, "nzs %l, %l" },
	{ xnewa, "newa %l, %l, %t" },
	{ xleaa, "leaa %l, %l, %l" },
	{ xleam, "leam %l, %l, %l" },
	{ xleaassoc, "leaassoc %l, %l, %l" },
	{ xlealistn, "lealistn %l, %l, %l" },
	{ xlengtha, "lengtha %l, %l" },
	{ xlengthl, "lengthl %l, %l" },
	{ xlengths, "lengths %l, %l" },
	{ xzero, "zero %l, %t" },
	{ xzeron, "zeron %l, %i" },
	{ xasttoa, "asttoa %l, %l" },
	{ xasttolist, "asttolist %l, %l" },
	{ xasttooption, "asttooption %l, %l" },
	{ xastparent, "astparent %l, %l, %t" },
	{ xadds, "adds %l, %l, %l" },
	{ xaddl, "addl %t, %l, %l, %l" },
	{ xatoi, "atoi %l, %l" },
	{ xatof, "atof %l, %l" },
	{ xitoa, "itoa %l, %l" },
	{ xbtoa, "btoa %l, %l" },
	{ xftoa, "ftoa %l, %l" },
	{ xftoi, "ftoi %l, %l" },
	{ xbtof, "btof %l, %l" },
	{ xitof, "itof %l, %l" },
	{ xreversel, "reversel %l, %l, %t" },
	{ xstringindex, "stringindex %l, %l, %l" },
	{ xslicea, "slicea %l, %l, %l, %l" },
	{ xslices, "slices %l, %l, %l, %l" },
	{ xges, "ges %l, %l, %l" },
	{ xgts, "gts %l, %l, %l" },
	{ xles, "les %l, %l, %l" },
	{ xlts, "lts %l, %l, %l" },
	{ xlisttoarray, "listtoarray %t, %l, %l" },
	{ xlistmap, "listtoarray %t, %l, %l, %p" },
	{ xarraytolist, "arraytolist %t, %l, %l" },
	{ xissubtype, "issubtype %l, %l, %l" },
	{ xcancastast, "cancastast %l, %t, %l" },
	{ xtagstring, "tagstring %l, %l, %t" },
	{ xnewcopy, "newcopy %l, %l, %t" },
	{ xnewmapiter, "newmapiter %l, %l" },
	{ xmapiternext, "mapiternext %i, %l" },
	{ xmapitertuple, "mapitertuple %l, %l" },
	{ xparse, "parse %l, %l, %l, %l" },
	{ xattr, "attr %l, %l, %p" },
	{ xnop, "nop" },
};

static Disa*
finddis(Code *cp)
{
	int i;
	
	for(i=0; i<nelem(distab); i++)
		if(distab[i].code == cp)
			return &distab[i];
	return nil;
}

void
checkdisasm(void)
{
	int i;
	
	for(i=0; i<nelem(nametab); i++)
		if(finddis(nametab[i].code) == nil)
			fprint(2, "disasm: missing fmt for %s\n", nametab[i].fmt);
}

int
locfmt(Fmt *fmt)
{
	Loc *l;

	l = va_arg(fmt->args, Loc*);
	if(l == nil)
		return fmtstrcpy(fmt, "nil");
	switch(l->mode){
	default:
		return fmtstrcpy(fmt, "???");
	case LocGlobal:
		return fmtprint(fmt, "%d(GP)", l->off);
	case LocGlobalIndirect:
		return fmtprint(fmt, "%d(%d(GP))", l->off1, l->off);
	case LocFrame:
		return fmtprint(fmt, "%d(FP)", l->off);
	case LocFrameIndirect:
		return fmtprint(fmt, "%d(%d(FP))", l->off1, l->off);
	case LocEscape:
		return fmtprint(fmt, "%d(EP.%d)", l->off, l->depth);
	case LocEscapeIndirect:
		return fmtprint(fmt, "%d(%d(EP.%d))", l->off1, l->off, l->depth);
	}
}


int
disasm1(char *s, char *es, uchar *ip, uchar *eip)
{
	Fmt fmt;
	union {
		Code **pc;
		int *ip;
		double *fp;
		void **vp;
		uchar *cp;
	} u;
	Code *code;
	Disa *dis;
	char *tmp;
	char *f;
	int i;
	int line;
	double d;
	void *v;
	Loc l;
	
	u.cp = ip;
	if(u.cp == eip)
		return 0;
	if(u.cp+sizeof(int)+sizeof(Code*) > eip)
		return -1;
	line = *u.ip++;
	code = *u.pc++;
	dis = finddis(code);
	if(dis == nil)
		return -1;
	
	fmtstrinit(&fmt);
	fmtprint(&fmt, "%s:%d  ", num2file(line>>16), line&0xffff);
	for(f=dis->fmt; *f; f++){
		if(*f != '%'){
			fmtrune(&fmt, *f);
			continue;
		}
		f++;
		switch(*f){
		case 0:
			fmtrune(&fmt, '%');
			f--;
			break;
		default:
			fmtrune(&fmt, '%');
			fmtrune(&fmt, *f);
			break;
		case 'r':	/* relative pc location */
			if(u.cp+sizeof(int) > eip)
				return -1;
			i = *u.ip++;
			fmtprint(&fmt, "%+#x [%p]", i, u.cp+i);
			break;
		case 'i':
			if(u.cp+sizeof(int) > eip)
				return -1;
			i = *u.ip++;
			fmtprint(&fmt, "%d", i);
			break;
		case 'f':
			if(u.cp+sizeof(double) > eip)
				return -1;
			d = *u.fp++;
			fmtprint(&fmt, "%.20g", d);
			break;
		case 'p':
			if(u.cp+sizeof(void*) > eip)
				return -1;
			v = *u.vp++;
			fmtprint(&fmt, "%p", v);
			break;
		case 's':
			if(u.cp+sizeof(void*) > eip)
				return -1;
			v = *u.vp++;
			fmtprint(&fmt, "%s", v);
			break;
		case 't':
			if(u.cp+sizeof(void*) > eip)
				return -1;
			v = *u.vp++;
			fmtprint(&fmt, "%T", v);
			break;
		case 'l':
			if(u.cp+sizeof(int) > eip)
				return -1;
			i = *u.ip++;
			if(i == ~0){
				fmtprint(&fmt, "z");
				break;
			}
			l.mode = i&7;
			l.off = i>>3;
			l.off1 = -1;
			l.depth = -1;
			if((l.mode&~LocIndirect) == LocEscape){
				l.depth = (i>>3)&0xFF;
				l.off = i>>11;
			}
			if(l.mode&LocIndirect){
				if(u.cp+sizeof(int) > eip)
					return -1;
				l.off1 = *u.ip++;
			}
			fmtprint(&fmt, "%&", &l);
			break;
		}
	}
	
	tmp = fmtstrflush(&fmt);
	utfecpy(s, es, tmp);
	free(tmp);
	return u.cp - ip;
}

void
disasm(uchar *i, uchar *ei)
{
	int n;
	char buf[100];
	
	while((n = disasm1(buf, buf+sizeof buf, i, ei)) > 0){
		fprint(2, "%p: %s\n", i, buf);
		i += n;
	}
	if(n < 0)
		fprint(2, "decoding error: %p\n", *(char**)i);
}

