/*
 * Zeta.  New simple interpreted language
 *   (because there weren't enough already!)
 */
/*
 * This grammar has:
 * 1 shift/reduce conflicts
 *   - (if else)
 * 1 reduce/reduce conflict
 *   - {}; is either an empty block or an empty map; resolved for block
 */
%{
#define NO_Y_TAB_H
#include "grammar.h"

#define listappend(x, y) listappend(asyntax, x, y)
#define listinsert(x, y) listinsert(asyntax, x, y)
#define list1(x)         list1(asyntax, x)
#define list2(x, y)      list2(asyntax, x, y)

#define MENTION(x)
#define YYLTYPE Line
#define YYLLOC_DEFAULT(Current, Rhs, N)          \
  if(N >= 1) Current = linerange(Rhs[1], Rhs[N]); else Current = ZL

// Return a line encompassing a thru b.
static Line
linerange(Line a, Line b)
{
	if(a.line == 0)
		return b;
	if(b.line == 0)
		return a;
	a.line1 = b.line1;
	a.byte1 = b.byte1;
	return a;
}

#define L(e) ({ typeof(e) _t = (e); _t->line = (yyloc); _t; })
#define L_(e) ({ typeof(e) _t = (e); _t->line_ = (yyloc); _t; })

static Pattern*
buildListPattern(List<Pattern*> *l)
{
	if(!l)
		return mkPatternEmptyList();
	else
		return mkPatternList(l->x, buildListPattern(l->next));
}

static void
parsetypedslot(Name slot, Name *name, Name *type)
{
	char *c;

	if((c = strchr(slot,':')) == nil)
		panic("bad typed slot");
	*name = nameofn(slot, c-slot);
	*type = nameof(c+2);
}

static GTerm*
parsegtermname(Name name)
{
	int len, end;
	Name sep;
	char *fst;

	len = strlen(name);
	end = name[len-1];
	if(end == '?')
		return mkGTermOpt(nameofn(name, len-1));
	if(end != '*' && end != '+')
		return mkGTermName(name);
	if(len > 3 && name[len-2] == ']'){
		fst = strchr(name, '[');
		assert(fst);
		sep  = nameofn(fst+1, len-3-(fst-name));
		name = nameofn(name, fst-name);
	}else{
		sep  = nil;
		name = nameofn(name, len-1);
	}
	if(end == '*')
		return mkGTermStar(name, sep);
	else
		return mkGTermPlus(name, sep);
}

%}

#tokenlist#

%pure_parser

%token	Eof Error
%token	<name>	LName LTermName LTermLBracket LTermLParen
%token	<name>	LSlot LTypedSlot LSlotExprStart
%token	<i>		LInteger
%token	<f>		LFloat
%token	<name>	LString LRegexp LToken

/*
 * Order of operations.  Similar to, but different from C.
 * In particular, the bitwise operators have been moved
 * below the booleans, so that a&4 == b needs no parentheses.
 *
 * Also some new operators: &&=, ||=, :=.
 *
 * The arrow operator for structure dereferencing is gone,
 * replaced by plain dot.
 */
%left	"->"
%left	"list" "ptr" "array" "map"

%left	","
%right	"=" "+=" "-=" "*=" "/=" "%=" "<<=" ">>=" "&=" "^=" "|=" "&&=" "||=" ":="
%left	"||"
%left	"&&"
%left	"==" "~~"
%left	"!=" "!~~"
%left	"<" ">" "<=" ">="
%left	"|"
%left	"^"
%left	"&"
%left	"<<" ">>"
%right "#"
%left	"+" "-"
%left	"*" "/" "%"
%left	Unary "!" "~" "!~" "new"
%left	"--" "++" "." "[" "]" "(" ")"
%left	LString
%left "{" "}"
%left "@"
%left HI

%%

toplist:
|	toplist top

top:
	stmt	{ dotop($1); showprompt(); }
|	Eof		{ return 0; }
|	"include" LString ";"  { doinclude($2); showprompt(); }
|	"require" LString ";"  { dorequire(@$, $2, 1); showprompt(); }
|	"restructure" name ";" { addrestructure(@$, $2); showprompt(); }

%type <stmt> stmt
stmt:
	zexpr ";"
		{ $$ = L_(new StmtExpr($1)); }
|	"const" expr ";"			
		{ StmtExpr *s = L_(new StmtExpr($2)); s->flags_ |= DeclareConst; $$ = s; }
|	name ":" type ";"			
		{ $$ = L_(new StmtDeclare($1, $3, ZList)); }
|	name ":" "<" namelist ">" type ";"			
		{ $$ = L_(new StmtDeclare($1, $6, $4)); }
|	"typedef" name ":" type ";"			
		{ $$ = L_(new StmtDeclareType($2, $4)); }
|	"for" "(" zexpr ";" zexpr ";" zexpr ")" stmt
		{ $$ = L_(new StmtFor($3, $5, $7, $9)); }
|	"for" "(" expr "in" expr ")" stmt
		{ $$ = L_(new StmtForList($3, $5, $7)); }
|	"if" "(" expr ")" stmt "else" stmt
		{ $$ = L_(new StmtIf($3, $5, $7)); }
|	"if" "(" expr ")" stmt
		{ $$ = L_(new StmtIf($3, $5, nil)); }
|	"while" "(" expr ")" stmt
		{ $$ = L_(new StmtWhile($3, $5)); }
|	"do" stmt "while" "(" expr ")" ";"
		{ $$ = L_(new StmtDoWhile($2, $5)); }
|	"return" zexpr ";"				
		{ $$ = L_(new StmtReturn($2)); }
|	"break" ";"						
		{ $$ = L_(new StmtBreak()); }
|	"continue" ";"					
		{ $$ = L_(new StmtContinue()); }
|	"{" zstmtlist "}"				
		{ $$ = L_(new StmtList($2)); }
|	"print" expr ";"
		{ $$ = L_(new StmtPrint($2)); }
|	"goto" name ";"
		{ $$ = L_(new StmtGoto($2)); }
|	"label" name ";"
		{ $$ = L_(new StmtLabel($2)); }
|	fnflags fn
		{ $$ = L_($2); $2->flags_ = $1; }
|	fnflags name "::" name ":" type ";"
		{
			if($1 != FnAttribute){
				fprint(2, "%L: illegal declaration\n", @1);
				YYERROR;
			}
			$$ = L_(new StmtDeclareAttribute(list1($2), $4, $6));
		}
|	fnflags name "." unprimed_parsename "::" name ":" type ";"
		{
			if($1 != FnAttribute){
				fprint(2, "%L: illegal declaration\n", @1);
				YYERROR;
			}
			$$ = L_(new StmtDeclareAttribute(listinsert($4, $2), $6, $8));
		}

%type <flags> fnflags
fnflags:
	"fn"                        { $$ = 0; }
|	"attribute"                 { $$ = FnAttribute; }
|	"extensible" "fn"           { $$ = FnExtensible; }
|	"extensible" "attribute"    { $$ = FnExtensible | FnAttribute; }
|	"extend" "fn"               { $$ = FnExtend; }
|	"extend" "attribute"        { $$ = FnExtend | FnAttribute; }

%type <stmtfndef> fn
fn:
	name "(" zfnarglist ")" zfnret "{" stmtlist "}"
	{ $$ = L_(new StmtFndef($1, mkfn(@1, $1, $3, $5, $7), ZList)); }
|	name "<" namelist ">" "(" zfnarglist ")" zfnret "{" stmtlist "}"
	{ $$ = L_(new StmtFndef($1, mkfn(@1, $1, $6, $8, $10), $3)); }

%type <expr> exprfn
exprfn:
	"fn" "(" zfnarglist ")" zfnret "{" stmtlist "}"
	{ $$ = L_(new ExprFn(mkfn(@1, Name::null, $3, $5, $7), ZList)); }
|	"fn" "<" namelist ">" "(" zfnarglist ")" zfnret "{" stmtlist "}"
	{ $$ = L_(new ExprFn(mkfn(@1, Name::null, $6, $8, $10), $3)); }

%type <fnarglist> fnarglist
fnarglist:
	fnarg					{ $$ = list1($1); }
|	fnarglist "," fnarg		{ $$ = listappend($1, $3); }

%type <fnarglist> zfnarglist
zfnarglist:					{ $$ = ZList; }
|	fnarglist

%type <type> zfnret
zfnret:						{ $$ = typevoid(); }
|	fnret

%type <type> fnret
fnret:
	":" type				{ $$ = $2; }

%type <fnarg> fnarg
fnarg:
	name ":" type			{ $$ = mkFnarg($1, $3); }

%type <stmtlist> stmtlist
stmtlist:
	stmt					{ $$ = list1($1); }
|	stmtlist stmt			{ $$ = listappend($1, $2); }

%type <stmtlist> zstmtlist
zstmtlist:					{ $$ = ZList; }
|	stmtlist				{ $$ = $1; }

/* mention @$ in a rule below so that bison generates yyloc-tracking code */

%type <expr> expr
expr:
	constant			{ MENTION(@$); $$ = $1; }
|	name				{ $$ = L_(new ExprName($1)); }
|	expr "+" expr			{ $$ = L_(new ExprAdd($1, $3)); }
|	expr "-" expr			{ $$ = L_(new ExprSub($1, $3)); }
|	expr "*" expr			{ $$ = L_(new ExprMul($1, $3)); }
|	expr "/" expr			{ $$ = L_(new ExprDiv($1, $3)); }
|	expr "%" expr			{ $$ = L_(new ExprMod($1, $3)); }
|	expr "&" expr			{ $$ = L_(new ExprAnd($1, $3)); }
|	expr "|" expr			{ $$ = L_(new ExprOr($1, $3)); }
|	expr "^" expr			{ $$ = L_(new ExprXor($1, $3)); }
|	expr "<<" expr			{ $$ = L_(new ExprLsh($1, $3)); }
|	expr ">>" expr			{ $$ = L_(new ExprRsh($1, $3)); }
|	expr "==" expr			{ $$ = L_(new ExprEqEq($1, $3)); }
|	expr "!=" expr			{ $$ = L_(new ExprNotEq($1, $3)); }
|	expr "&&" expr			{ $$ = L_(new ExprAndAnd($1, $3)); }
|	expr "||" expr			{ $$ = L_(new ExprOrOr($1, $3)); }
|	expr "=" expr			{ $$ = L_(new ExprEq($1, $3)); }
|	expr ":=" expr			{ $$ = L_(new ExprDeclEq($1, nil, $3)); }
|	name ":" type "=" expr	{ $$ = L_(new ExprDeclEq(L_(new ExprName($1)), $3, $5)); }
|	expr "+=" expr			{ $$ = L_(new ExprAddEq($1, $3)); }
|	expr "-=" expr			{ $$ = L_(new ExprSubEq($1, $3)); }
|	expr "*=" expr			{ $$ = L_(new ExprMulEq($1, $3)); }
|	expr "/=" expr			{ $$ = L_(new ExprDivEq($1, $3)); }
|	expr "%=" expr			{ $$ = L_(new ExprModEq($1, $3)); }
|	expr "&=" expr			{ $$ = L_(new ExprAndEq($1, $3)); }
|	expr "|=" expr			{ $$ = L_(new ExprOrEq($1, $3)); }
|	expr "^=" expr			{ $$ = L_(new ExprXorEq($1, $3)); }
|	expr "<<=" expr			{ $$ = L_(new ExprLshEq($1, $3)); }
|	expr ">>=" expr			{ $$ = L_(new ExprRshEq($1, $3)); }
|	"(" zexprlist ")"		{ $$ = L_(new ExprTuple(*$2)); }
|	expr "(" zexprlist ")"		{ $$ = L_(new ExprCall($1, *$3)); }
|	expr "(" initlist ")"		{ $$ = L_(new ExprInit($1, *$3)); }
|	expr "[" expr "]"		{ $$ = L_(new ExprArray($1, $3)); }
|	expr "[" expr "," zexpr "]"	{ $$ = L_(new ExprSlice($1, $3, $5)); }
|	expr "<" expr			{ $$ = L_(new ExprLt($1, $3)); }
|	expr "<=" expr			{ $$ = L_(new ExprLe($1, $3)); }
|	expr ">" expr			{ $$ = L_(new ExprGt($1, $3)); }
|	expr ">=" expr			{ $$ = L_(new ExprGe($1, $3)); }
|	"~" expr			{ $$ = L_(new ExprTwid($2)); }
|	"+" expr %prec Unary	{ $$ = L_(new ExprPos($2)); }
|	"-" expr %prec Unary	{ $$ = L_(new ExprNeg($2)); }
|	"*" expr %prec Unary	{ $$ = L_(new ExprStar($2)); }
|	"&" expr %prec Unary	{ $$ = L_(new ExprAddr($2)); }
|	"!" expr			{ $$ = L_(new ExprNot($2)); }
|	exprfn
|	expr "." name			{ $$ = L_(new ExprDot($1, $3)); }
|	"new" expr			{ $$ = L_(new ExprNew($2)); }
|	"`" parsename "(" exprlist ")"	{ $$ = L_(new ExprParseString($2, *$4)); }
|	"`" parseexpr			{ $$ = L_(new ExprRestructure($2, nil)); }
|	"`" parseexpr "@" expr	{ $$ = L_(new ExprRestructure($2, $4)); }
|	expr "~" parseexpr		{ $$ = L_(new ExprDestructure($1, $3)); }
|	expr "~" parsename		{ $$ = L_(new ExprDestructType($1, $3)); }
|	expr "!~" parseexpr		{ $$ = L_(new ExprNot(L_(new ExprDestructure($1, $3)))); }
|	expr "!~" parsename		{ $$ = L_(new ExprNot(L_(new ExprDestructType($1, $3)))); }
|	expr "~~" expr			{ $$ = L_(new ExprSameStructure($1, $3)); }
|	expr "!~~" expr			{ $$ = L_(new ExprNot(new ExprSameStructure($1, $3))); }
|	"bool" "(" expr ")"		{ $$ = L_(new ExprConvert($3, typebool())); }
|	"int" "(" expr ")"		{ $$ = L_(new ExprConvert($3, typeint())); }
|	"float" "(" expr ")"		{ $$ = L_(new ExprConvert($3, typefloat())); }
|	"string" "(" expr ")"		{ $$ = L_(new ExprConvert($3, typestring())); }
|	"list" "(" expr ")"		{ $$ = L_(new ExprConvert($3, typelist(nil))); }
|	"array" "(" expr ")"		{ $$ = L_(new ExprConvert($3, typearray(nil))); }
|	"option" "(" expr ")"		{ $$ = L_(new ExprConvert($3, typeptr(nil))); }
|	"length" "(" expr ")"		{ $$ = L_(new ExprLength($3)); }
|	"tag" "(" expr ")"		{ $$ = L_(new ExprTag($3)); }
|	"[" zexprlist "]"	{ $$ = L_(new ExprMkListOrArray(*$2)); }
|	"list" "[" zexprlist "]"	{ $$ = L_(new ExprMkList(*$3)); }
|	"array" "[" zexprlist "]"	{ $$ = L_(new ExprMkArray(*$3)); }
|	"{" zmaplist "}"		{ $$ = L_(new ExprMkMap(*$2)); }
|	expr "#" expr			{ $$ = L_(new ExprListCons($1, $3)); }
|	expr "++"			{ $$ = L_(new ExprPostInc($1)); }
|	expr "--"			{ $$ = L_(new ExprPostDec($1)); }
|	"++" expr			{ $$ = L_(new ExprPreInc($2)); }
|	"--" expr			{ $$ = L_(new ExprPreDec($2)); }
|	"reverse" "(" expr ")"		{ $$ = L_(new ExprReverse($3)); }
|	"astparent" "(" expr "," parsename ")"	{ $$ = L_(new ExprAstParent($3, $5)); }

%type <initfield> init
init:	"." name "=" expr	{ $$ = new ExprInitField($2, $4); }

%type <initfields> initlist
initlist: init			{ $$ = new vector<ExprInitField*>; $$->push_back($1); }
|	initlist "," init	{ $$ = $1; $$->push_back($3); }

%type <exprmap> map
map:	expr "=>" expr		{ $$ = new ExprMap($1, $3); }

%type <exprmaps> zmaplist
zmaplist: 	{ $$ = new vector<ExprMap*>; }
|	maplist	{ $$ = $1; }

%type <exprmaps> maplist
maplist: map	{ $$ = new vector<ExprMap*>; $$->push_back($1); }
|	maplist "," map	{ $$ = $1; $$->push_back($3); }

%type <expr> constant
constant:
	"nil"					{ $$ = L_(new ExprNil()); }
|	"true"					{ $$ = L_(new ExprBool(1)); }
|	"false"					{ $$ = L_(new ExprBool(0)); }
|	LInteger				{ $$ = L_(new ExprInteger($1)); }
|	LFloat					{ $$ = L_(new ExprFloat($1)); }
|	LString					{ $$ = L_(new ExprString($1, $<len>1)); }

%type <exprs> exprlist
exprlist:
	expr					{ $$ = new vector<Expr*>;  $$->push_back($1); }
|	exprlist "," expr		{ $$ = $1; $$->push_back($3); }

%type <expr> zexpr
zexpr:						{ $$ = nil; @$ = ZL; }
|	expr

%type <exprs> zexprlist
zexprlist:					{ $$ = new vector<Expr*>; }
|	exprlist

%type <type> type
type:
	"bool"					{ $$ = typebool(); }
|	"int"					{ $$ = typeint(); }
|	"string"				{ $$ = typestring(); }
|	"void"					{ $$ = typevoid(); }
|	"float"					{ $$ = typefloat(); }
|	name					{ $$ = mkTypeName($1, ZList); }
|	name "<" typelist ">"	{ $$ = mkTypeName($1, $3); }
|	"list" type				{ $$ = typelist($2); }
|	"array" type				{ $$ = typearray($2); } 
|	"map" "(" type "," type ")"	{ $$ = typemap($3, $5); }
|	type "->" type 			{ $$ = typefn2($1, $3); }
|	"(" ztypelist ")"		{ $$ = typetuple($2); }
|	"ptr" type				{ $$ = typeptr($2); }
|	type "." termname		{ $$ = typesubname($1, $3); }
|	name "?"				{ $$ = resolvegramwildcard(@1, $1); }

%type <typelist> typelist
typelist:
	type					{ $$ = list1($1); }
|	typelist "," type		{ $$ = listappend($1, $3); }

%type <typelist> ztypelist
ztypelist:					{ $$ = ZList; }
|	typelist

%type <name> name
name:	LName	{ $$ = $1; @$ = @1; }

%type <namelist> namelist
namelist:
	name                 { $$ = list1($1); }
|	namelist "," name    { $$ = listappend($1, $3); }

%type <name> termname
termname:	Intermname LTermName Notintermname
		{ $$ = $2; @$ = @2; }

stmt:
	structdecl
		{ $$ = $1; }
|	"extensible" structdecl
		{ $$ = L_($2); $2->flags_ = FnExtensible; }
|	"extend" structdecl
		{ $$ = L_($2); $2->flags_ = FnExtend; }
|	"discriminated" structdecl
		{ $$ = L_($2); $2->discriminated_ = 1; }
|	"extensible" "discriminated" structdecl
		{ $$ = L_($3); $3->discriminated_ = 1; $3->flags_ = FnExtensible; }

%type <stmtstruct> structdecl
structdecl:
	"struct" name "{" decllist "}" ";"
		{ $$ = L_(new StmtStruct($2, $4, ZList)); }
|	"struct" name "<" namelist ">" "{" decllist "}" ";"
		{ $$ = L_(new StmtStruct($2, $7, $4)); }

%type <structelemlist> decllist
decllist:	{ $$ = ZList; }
|	decllist structelem	{ $$ = listappend($1, $2); }

%type <structelem> structelem
structelem:
	fnarg ";"	{ $$ = L(mkStructelemDecl($1)); }
|	"tag" termname "{" decllist "}" ";"
		{ $$ = L(mkStructelemTag($2, Name::null, $4)); }
|	"tag" termname ":" termname "{" decllist "}" ";"
		{ $$ = L(mkStructelemTag($2, $4, $6)); }

/* Grammar definitions */
Ingrammar: { ingrammar= 1; }
Notingrammar: { ingrammar= 0; }
top:
	grammardecl "{" Ingrammar gramlist Notingrammar "}"
		{
			$1->grams_ = $4;
			dotop(L_($1));
		}

%type <stmtgrammar> grammardecl
grammardecl:
	"grammar" name gextends
		{ $$ = L_(new StmtGrammar($2, $3, nil)); }
|	"extensible" "grammar" name
		{ $$ = L_(new StmtGrammar($3, nil, nil)); }
|	"extend" "grammar" name
		{ $$ = L_(new StmtGrammar(Name::null, list1($3), nil)); }

%type <namelist> gextends
gextends:
		{ $$ = ZList; }
|	"extends" namelist
		{ $$ = $2; }

%type <pgramlist> gramlist
gramlist:
		{ $$ = ZList; }
|	gramlist gram
		{ $$ = listconcat($1, $2); }

%type <pgramlist> gram
gram:
	name ":" Intermname grulelist ";" Notintermname
		{
			List<PGram*> *l;
			for(l=$4; l; l=l->next){
				PGram *g = l->x;
				g->left = mkGTermName($1);
			}
			$$ = $4;
		}
|	"%ignore" Intermname gtermlist Notintermname ";"
		{ $$ = list1(mkPGramIgnore(@$, $3)); }
|	"%share" gtermnamelist ";"
		{ $$ = list1(mkPGramShare(@$, $2)); }
|	"%view" name name name name "(" gproplist ")" ";"
		{
			if(strcmp($3, "of") != 0 && strcmp($5, "where") != 0)
				YYERROR;
			$$ = list1(mkPGramView(@$, $2, $4, $7));
		}
|	assoc preclist ";"
		{ $$ = list1(mkPGramAssoc(@$, $1, $2)); }
|	"%priority" precorder ";"
		{ $$ = list1(mkPGramPriority(@$, $2)); }
|	"%prefer" precorder ";"
		{ $$ = list1(mkPGramPrefer(@$, $2)); }

%type <pgramlist> grulelist
grulelist:
	grule
		{ $$ = list1($1); }
|	grulelist "|" grule
		{ $$ = listappend($1, $3); }

%type <pgram> grule
grule:
	zgrulename gtermlist prec
		{ $$ = mkPGramRule(@$, $1, nil, $2, $3); }

%type <name> zgrulename
zgrulename:
		{ @$ = ZL; $$ = nil; }
| "<" Notintermname name Intermname ">"
		{ @$ = @$; $$ = $3; }

%type <gtermlist> gtermlist
gtermlist:
		{ $$ = ZList; }
|	gtermlist gterm
		{ $$ = listappend($1, $2); }

%type <gterm> gterm
gterm:
	termname Intermname
		{ $$ = parsegtermname($1); }
|	LString
		{ $$ = mkGTermString($1, $<len>1); }
|	LRegexp
		{ $$ = mkGTermRegexp($1, $<len>1, $<flags>1); }

%type <name> prec
prec:                                       { $$ = Name::null; }
|	"[" Notintermname precitem Intermname "]" { $$ = $3; }

%type <assoc> assoc
assoc:
	"%left"                 { $$ = CfgAssocLeft; }
|	"%right"                { $$ = CfgAssocRight; }
|	"%nonassoc"             { $$ = CfgAssocNone; }

%type <namelist> precorder
precorder:
	precitem                { $$ = list1($1); }
|	precorder ">" precitem  { $$ = listappend($1, $3); }

%type <namelist> preclist
preclist:
	precitem                { $$ = list1($1); }
|	preclist precitem       { $$ = listappend($1, $2); }

%type <name> precitem
precitem:
	name                    { $$ = $1; }
|	LString                 { $$ = nameof(cescapestring($1, -1, '"')); }

%type <namelistlist> gtermnamelist
gtermnamelist:            { $$ = ZList; }
|	gtermnamelist gtermname { $$ = listappend($1, $2); }

%type <namelist> gtermname
gtermname:
	name                    { $$ = list1($1); }
|	name "." name           { $$ = list2($1, $3); }

%type <namelistlist> gproplist
gproplist:
	gprop                   { $$ = list1($1); }
|	gproplist "," gprop     { $$ = listappend($1, $3); }

%type <namelist> gprop
gprop:
	name "=" propnamelist   { $$ = listinsert($3, $1); }

%type <namelist> propnamelist
propnamelist:
	name                          { $$ = list1($1); }
|	"{" termname "}"              { $$ = list1($2); }
|	propnamelist name             { $$ = listappend($1, $2); }
|	propnamelist "{" termname "}" { $$ = listappend($1, $3); }

/* Pattern matching */
Backtick: { inbacktick = 1; }
Nobacktick: { inbacktick = 0; }
Intermname:	{intermname = 1;}
Notintermname:	{intermname = 0;}

%type <exprparse> parseexpr
parseexpr:
	parsename "{" Backtick tokenlist Nobacktick "}"
		{ $$ = L_(new ExprParse($1, $4)); }
|	Intermname LTermLBracket Notintermname Backtick tokenlist Nobacktick "}"
		{ $$ = L_(new ExprParse(nil, $5)); }

%type <namelist> unprimed_parsename
unprimed_parsename:
	termname
		{ $$ = list1($1); }
|	parsename "." termname
		{ $$ = listappend($1, $3); }

%type <namelist> parsename
parsename:
	termname
		{
			/* prime the type so it exists before returning the parsename */
			if(strchr($1,'?') != nil)
				resolvegramwildcard(@1, nameofn($1, strlen($1)-1));
			$$ = list1($1);
		}
|	parsename "." termname
		{ $$ = listappend($1, $3); }

%type <ptokenlist> tokenlist
tokenlist:
		{ $$ = ZList; }
|	tokenlist LToken
		{ $$ = listappend($$, mkPTokenString($2, $<len>2)); }
|	tokenlist LSlot
		{
			PToken *gt;
			ExprName *en;
			gt = mkPTokenExpr(L_(en = new ExprName($2)));
			gt->slot = en->name_;
			$$ = listappend($$, gt);
		}
|	tokenlist LTypedSlot
		{
			PToken *gt;
			ExprName *en;
			gt = mkPTokenExpr(L_(en = new ExprName(Name::null)));
			parsetypedslot($2, &en->name_, &gt->symname);
			gt->slot = en->name_;
			$$ = listappend($$, gt);
		}
|	tokenlist LSlotExprStart Nobacktick expr ")" Backtick
		{ $$ = listappend($$, mkPTokenExpr($4)); }

stmt:
	"switch" "(" exprlist ")" "{" clauselist "}"
		{
			Expr *cond;
			if($3->size() == 1)
				cond = (*$3)[0];
			else
				cond = new ExprTuple(*$3);
			$$ = L_(new StmtCase(cond, $6));
		} 

%type <clauselist> clauselist
clauselist:
	clause			{ $$ = list1($1); }
|	clauselist clause	{ $$ = listappend($1, $2); }

%type <clause> clause
clause:
	"case" pattern ":" stmtlist	{ $$ = mkClause($2, new StmtList($4)); }

%type <pattern> pattern
pattern:
	name pattern	%prec HI
		{ $$ = L(mkPatternTag(list1($1), $2)); }
|	"(" zpatlist ")"
		{ $$ = L(mkPatternDtor($2)); }
|	"[" zpatlist "]"
		{ $$ = L(buildListPattern($2)); }
|	pattern "#" pattern
		{ $$ = L(mkPatternList($1, $3)); }
|	"~" parsename
		{ $$ = L(mkPatternTag($2, mkPatternAny())); }
|	"~" parsename name
		{
			if(strcmp($3, "_") == 0)
				$$ = L(mkPatternTag($2, mkPatternAny()));
			else
				$$ = L(mkPatternTag($2, mkPatternLit($3, 0)));
		}
|	"~" parseexpr
		{ $$ = L(mkPatternParse($2, Name::null)); }
|	"~" parseexpr name
		{ $$ = L(mkPatternParse($2, $3)); }
|	constant
		{ $$ = L(mkPatternConst($1)); }
|	name
		{
			if(strcmp($1, "_") == 0)
				$$ = mkPatternAny();
			else
				$$ = mkPatternLit($1, 0);
			$$ = L($$);
		}
|	"*" name
		{ $$ = L(mkPatternLit($2, 1)); }
|	pattern "||" pattern
		{ $$ = L(mkPatternDisjunct($1, $3)); }

%type <patternlist> patlist
patlist:
	pattern			{ $$ = list1($1); }
|	patlist "," pattern	{ $$ = listappend($1, $3); }

%type <patternlist> zpatlist
zpatlist:			{ $$ = ZList; }
|	patlist


/* Canonicalization */
stmt:
	"canonicalize" parsename ";"	{ $$ = L_(new StmtCanonicalize($2)); }

