/*
 * CFG CfgGram<void> management.
 */

#include "u.h"
#include "utf.h"
#include "fmt.h"
#include "name.h"
#include "arena.h"
#include "emalloc.h"
#include "list2.h"
#include "cfg.h"
#include "bitset.h"
#include "panic.h"

/*
 * Allocate, initialize a new grammar.
 */
CfgGram<void>*
_cfgnewgram(int sz)
{
	CfgGram<void> *g;

	if(sz < sizeof *g)
		panic("cfgnewgram: too small (%d)", sz);
	g = (CfgGram<void>*)emalloc(sz);
	g->valsize = -1;
	g->eofsym = cfgnewsym(g, "EOF", nil);
	return g;
}

/*
 * Free a grammar.
 */
void
_cfgfreegram(CfgGram<void> *g)
{
	int i;
	CfgRule<void> *r;
	CfgSym<void> *s;

	if(g->slrfreestate){
		for(i=0; i<g->nslrstate; i++)
			g->slrfreestate(g->slrstate[i]);
		efree(g->slrstate);
	}
	if(g->glrfreestate){
		for(i=0; i<g->nglrstate; i++)
			g->glrfreestate(g->glrstate[i]);
		efree(g->glrstate);
	}

	for(i=0; i<g->nsym; i++){
		s = g->sym[i];
		freebitset(s->first);
		efree(s->name);
		efree(s->type);
		efree(s->rule);
		efree(s);
	}
	efree(g->sym);
	
	for(i=0; i<g->nrule; i++){
		r = g->rule[i];
		efree(r->right);
		efree(r);
	}
	efree(g->rule);

	efree(g);
}

