/*
 * Generalized LR.
 */
struct GlrStack;
struct GlrPath;
struct GlrEdge;
template<typename T> struct CfgSym;
template<typename T> struct CfgRule;

template<typename T>
struct GlrToken
{
	CfgSym<T>	*sym;
	char val[1];
};

struct GlrEdge
{
	GlrEdge	*sib;
	GlrStack	*down;
	CfgSym<void>	*sym;
	char	val[1];
};

template<typename T>
struct GlrParse
{
	CfgGram<T>	*g;
	uint	seqnum;
	uint	ipos;
	GlrStack	**top;
	uint	ntop;
	uint	mtop;
	List<GlrToken<T>*>	*toks;
	T	*rtokval;
	T	*tmpval;
	GlrPath	**heap;
	uint	nheap;
	uint	mheap;
	Name file;
	int	line;
	int	col;
	void	*aux;
	void	(*merge)(CfgSym<T>*, T*, T*);
	void	(*dup)(CfgSym<T>*, T*, T*);
	void	(*free)(CfgSym<T>*, T*);
	List<GlrToken<T>*>	*freetoks;
};

struct GlrPath 
{
	int	leftpos;
	CfgRule<void>	*rule;
	GlrStack	*sp;
	GlrEdge	*edge[1];
};

struct GlrStack
{
	int		ipos;
	SlrState	*state;
	GlrEdge	*edge;
	int	ref;
};

typedef List<GlrToken<void>*> *GlrLexer(GlrParse<void>*, CfgGram<void>*, void*);

GlrParse<void>*	_glrnewparse(CfgGram<void>*);
void*		_glrparse(CfgGram<void> *g, CfgSym<void> *start, GlrParse<void> *p0, GlrLexer *yylex, void *arg);
void		_glrprintstate(GlrParse<void> *p);
GlrToken<void>*	_glrnewtoken(GlrParse<void>*, CfgSym<void>*);
void		_glrfreetoken(GlrParse<void>*, GlrToken<void>*);

template<typename T, typename State>
T*
glrparse(CfgGram<T> *g, CfgSym<T> *start, GlrParse<T> *p0, 
	List<GlrToken<T>*>* (*yylex)(GlrParse<T>*, CfgGram<T>*, State*),
	State *arg) 
{
	return (T*)_glrparse((CfgGram<void>*)g, (CfgSym<void>*)start, (GlrParse<void>*)p0, 
		(GlrLexer*)yylex, (void*)arg);
}

template<typename T>
GlrParse<T>*
glrnewparse(CfgGram<T> *g)
{
	return (GlrParse<T>*)_glrnewparse((CfgGram<void>*)g);
}

template<typename T>
GlrToken<T>*
glrnewtoken(GlrParse<T> *p, CfgSym<T> *s)
{
	return (GlrToken<T>*)_glrnewtoken((GlrParse<void>*)p, (CfgSym<void>*)s);
}

template<typename T>
void
glrfreetoken(GlrParse<T> *g, GlrToken<T> *t)
{
	_glrfreetoken((GlrParse<void>*)g, (GlrToken<void>*)t);
}

template<typename T>
void
glrprintstate(GlrParse<T> *p)
{
	_glrprintstate((GlrParse<void>*)p);
}

