#include "u.h"
#include "utf.h"
#include "fmt.h"
#include "name.h"
#include "emalloc.h"
#include "arena.h"
#include "list2.h"
#include "cfg.h"
#include "slr.h"
#include "bitset.h"

enum {
	BigHammer = 0	// Whether to clear entire SLR state at grammar change
};

int
slrcompile(CfgGram<void> *g)
{
	int i;

	if(g->dirty == 0)
		return 0;

	cfgcompile(g);

	if(BigHammer){
		for(i=0; i<g->nslrstate; i++)
			slrfreestate(g->slrstate[i]);
		efree(g->slrstate);
		g->slrfreestate = slrfreestate;
		g->slrstate = nil;
		g->nslrstate = 0;
		memset(g->slrstatehash, 0, sizeof g->slrstatehash);
	}

	g->seqnum++;
	g->dirty = 0;
	return 1;
}

int
slrcompilestates(CfgGram<void> *g)
{
	int i, j;
	SlrState *z;
	
	slrcompile(g);

	if(slrinitstate(g, nil) == nil)
		return -1;
	for(i=0; i<g->nslrstate; i++){
		z = g->slrstate[i];
		if(!z->compiled)
			slrcompilestate(g, z);
		for(j=0; j<z->nshift; j++)
			if(!z->shift[j].slrstate)
				z->shift[j].slrstate = slrcomputeshift(g, z, z->shift[j].sym);
	}
	return 0;
}

int
slrcompilestate(CfgGram<void> *g, SlrState *z)
{
	int i, n;
	SlrRulePos *p;
	Bitset *set;

assert(!z->compiled);
//	g->stats.slrcompilestate++;
	if(z->shift || z->nshift){
		efree(z->shift);
		z->shift = nil;
		z->nshift = 0;
	}
	if(z->reduce || z->nreduce){
		efree(z->reduce);
		z->reduce = nil;
		z->nreduce = 0;
	}

	/*
	 * Compute shift set
	 */
	set = newbitset(g->nsym);
	if(set == nil)
		return -1;
	for(i=0; i<z->npos; i++){
		p = &z->pos[i];
		if(p->i < p->r->nright)
			addbit(set, p->r->right[p->i]->n);
	}
	n = countbitset(set);
	z->shift = (SlrShift*)emalloc(n*sizeof(z->shift[0]));
	if(z->shift == nil)
		return -1;
	z->nshift = n;
	n = 0;
	for(i=0; i<g->nsym; i++){
		if(inbitset(set, i)){
			z->shift[n].sym = g->sym[i];
			z->shift[n].slrstate = nil;
			n++;
		}
	}
	freebitset(set);

	/*
	 * Compute reduce set
	 */
	n = 0;
	for(i=0; i<z->npos; i++){
		p = &z->pos[i];
		if(p->i == p->r->nright)
			n++;
	}
	z->reduce = (CfgRule<void>**)emalloc(n*sizeof z->reduce[0]);
	if(z->reduce == nil)
		return -1;
	z->nreduce = n;
	n = 0;
	for(i=0; i<z->npos; i++){
		p = &z->pos[i];
		if(p->i == p->r->nright)
			z->reduce[n++] = p->r;
	}

	z->compiled = 1;
	return 0;
}

void
mkallparseme(CfgGram<void> *g)
{
	int i;
	CfgSym<void> *s, *s0;

	for(i=0; i<g->nsym; i++){
		s = g->sym[i];
		if(s->parseme == nil && s->name[0] != '$'){
			s0 = cfgnewsym(g, nameprint("$%s", s->name), nil);
			cfgrulesym(g, (CfgAction(void)*)nil, (CfgPrec<void>*)nil, s0, s, cfglooksym(g, "EOF"), nil);
			s->parseme = s0->rule[0];
			s->parseme->toplevel = 1;
		}
	}
}

/*
 * on-the-fly compilation
 */
SlrState*
slrinitstate(CfgGram<void> *g, CfgSym<void> *start)
{
	SlrState *z;
	CfgSym<void> *s, *s0;
	
	if(start){
		s = start;
	}else{
		if(g->nrule == 0){
		//	werrstr("no rules");
			return nil;
		}
		s = g->rule[0]->left;
	}
	
	if(s->parseme == nil){
		s0 = cfgnewsym(g, nameprint("$%s", s->name), nil);
		cfgrulesym(g, (CfgAction(void)*)nil, (CfgPrec<void>*)nil, s0, s, cfglooksym(g, "EOF"), nil);
		s->parseme = s0->rule[0];
		s->parseme->toplevel = 1;
		mkallparseme(g);	// avoid lots of extra compiling
		slrcompile(g);
	}
	
	z = (SlrState*)emalloc(sizeof *z);
	z->g = g;
	z->pos = (SlrRulePos*)emalloc(32*sizeof z->pos[0]);
	z->pos[0].r = s->parseme;
	z->npos = 1;
//	fprint(2, "state0 %s(%d; %d) %lY\n", s->name, s->nrule, z->npos, z);
	z = slrcanonstate(g, z);
//	fprint(2, "canonstate %lY\n", z);
	return z;
}

