#include "u.h"
#include "utf.h"
#include "fmt.h"
#include "name.h"
#include "emalloc.h"
#include "arena.h"
#include "list2.h"
#include "cfg.h"
#include "slr.h"

static void
die(char *fmt, ...)
{
	va_list arg;
	
	fprint(2, "fatal: ");
	va_start(arg, fmt);
	vfprint(2, fmt, arg);
	fprint(2, "\n");
	exit(1);
}

SlrParse*
slrnewparse(void)
{
	return (SlrParse*)emalloc(sizeof(SlrParse));
}

static SlrParse*
slrparseinit(CfgGram<void> *g, SlrParse *p, CfgSym<void> *start)
{
	int i;
	char *val;
	
	if(p == nil)
		p = slrnewparse();
	p->stack = (SlrStack*)emalloc(100*sizeof p->stack[0]);
	p->estack = p->stack+100;
	p->sp = p->stack;
	val = (char*)emalloc((100+2)*g->valsize);
	p->tokval = val;
	val += g->valsize;
	p->tmpval = val;
	val += g->valsize;
	for(i=0; i<100; i++)
		p->stack[i].val = val+i*g->valsize;
	p->tok = nil;
	
	if(slrinitstate(g, start) == nil)
		return nil;
	slrcompile(g);
	
	p->sp->slrstate = slrinitstate(g, start);
	p->sp++;
	p->seqnum = g->seqnum;

//fprint(2, "initstate: %l@\n", p->sp[-1].slrstate);
	return p;
}

/*
 * Figure out what to do next, given current parse state and lookahead token.
 */
static int
action(CfgGram<void> *g, SlrParse *p, SlrState *z, SlrState **nz, CfgRule<void> **r, CfgSym<void> *tok)
{
	*nz = nil;
	*r = nil;
	if(!z->compiled)
		slrcompilestate(g, z);
	if(z->nshift==0 && z->nreduce==1){
		*r = z->reduce[0];
		return 1;
	}
	if(tok == nil)
		return 0;
	slrcanshift(g, z, tok, nz);
	slrcanreduce(g, p, z, tok, r);
	return 1;
}

/*
 * Recompute the parse states on the stack by re-shifting
 * if the grammar has been changed.
 */
static void
reshift(CfgGram<void> *g, CfgSym<void> *start, SlrParse *p)
{
	SlrStack *sp;
	
	if(!g->dirty && g->seqnum == p->seqnum)
		return;
fprint(2, "dirty %d seqnum %d %d\n", g->dirty, g->seqnum, p->seqnum);
	slrcompile(g);
	p->stack->slrstate = slrinitstate(g, start);
	for(sp=p->stack+1; sp<p->sp; sp++)
		if(!slrcanshift(g, (sp-1)->slrstate, sp->sym, &sp->slrstate))
			die("lost parse in reshift at %s", sp->sym->name);
	p->seqnum = g->seqnum;
}

/*
 * Parse
 */
void*
slrparse(CfgGram<void> *g, CfgSym<void> *start, SlrParse *p0, SlrLexer *slrlex, void *arg)
{
	SlrParse *p;
	SlrStack *sp;
	SlrState *nextstate;
	CfgRule<void> *rule;
	
//	fprint(2, "slrparse: start %s; valsize %d\n", start, g->valsize);
	p = slrparseinit(g, p0, start);
	if(p == nil)
		return nil;

	for(;;){
		if(p->sp >= p->estack)
			die("stack overflow");
//		fprint(2, "slrparse: [%s] %@\n", p->tok ? p->tok->name : "", p->sp[-1].slrstate);
		if(!action(g, p, p->sp[-1].slrstate, &nextstate, &rule, p->tok)){
			p->tok = slrlex(g, p->tokval, arg);
			reshift(g, start, p);
//			fprint(2, "slrparse: [%s] %@\n", p->tok ? p->tok->name : "", p->sp[-1].slrstate);
			action(g, p, p->sp[-1].slrstate, &nextstate, &rule, p->tok);
		}
		die("slrparse depends on obsolete rule->prec");
		if(nextstate && (!rule /*|| p->tok->prec >= rule->prec*/)){
			/* shift */
			sp = p->sp++;
			if(p->sp >= p->estack)
				die("stack overflow");
			sp->sym = p->tok;
			sp->slrstate = nextstate;
			// *(CfgSym<void>**)p->tokval = p->tok;
			memmove(sp->val, p->tokval, g->valsize);
			p->tok = nil;
		}else if(rule){
			/* reduce */
			if(rule->toplevel){
				p->sp -= 2;	/* start EOF */
				return p->sp->val;
			}
			p->sp -= rule->nright;
			memmove(p->tmpval, p->sp->val, g->valsize);
			if(rule->fn){
				rule->fn(g, rule, p->tmpval, p->sp->val);
				reshift(g, start, p);
			}
			// *(CfgSym<void>**)p->tmpval = rule->left;
			memmove(p->sp->val, p->tmpval, g->valsize);
			p->sp->sym = rule->left;
			if(!slrcanshift(g, (p->sp-1)->slrstate, p->sp->sym, &p->sp->slrstate))
				die("lost parse");
			p->sp++;
		}else{
			for(sp=p->stack+1; sp<p->sp; sp++)
				print("%s ", sp->sym->name);
			if(p->tok)
				print("%s\n", p->tok->name);
			else
				print("\n");
			die("parse error");
		}
	}	
}
