/*
 * SLR parsing
 */

typedef struct SlrRulePos SlrRulePos;
typedef struct SlrShift SlrShift;
typedef struct SlrState SlrState;
typedef struct SlrStack SlrStack;
typedef struct SlrParse SlrParse;

/* single position in a rule */
struct SlrRulePos
{
	CfgRule<void>	*r;
	int	i;
};

/*
 * Simple LR
 */
struct SlrShift
{
	CfgSym<void>	*sym;
	SlrState	*slrstate;
};

/* a single SLR state */
struct SlrState
{
	int	n;		/* state number */
	CfgGram<void>	*g;		/* grammar */
	SlrRulePos	*pos;		/* list of positions */
	int	npos;
	int	compiled;	/* are shift and reduce filled out? */
	SlrShift	*shift;		/* computed shifts */
	int	nshift;
	CfgRule<void>	**reduce;	/* computed reductions */
	int	nreduce;
	SlrState	*next;		/* in hash table */
	
	SlrState	*newer;		/* newer version of same state */
	
	uint	synctime;	/* last sync time */
};


struct SlrStack
{
	SlrState	*slrstate;
	CfgSym<void>	*sym;
	void		*val;
};

struct SlrParse
{
	CfgGram<void>	*g;				/* grammar */
	uint		seqnum;			/* grammar sequence number */
	SlrStack	*stack;
	SlrStack	*estack;
	SlrStack	*sp;
	CfgSym<void>	*tok;				/* lookahead token */
	void		*tokval;
	void		*tmpval;

	void	*aux;				/* for caller use */
};
typedef CfgSym<void>*	SlrLexer(CfgGram<void>*, void*, void*);

SlrParse*	slrnewparse(void);
void*		slrparse(CfgGram<void> *g, CfgSym<void> *start, SlrParse *p0, SlrLexer *slrlex, void *arg);
// XXX slrfreeparse
int		slrstatefmt(Fmt *fmt);

// Internal use only.
SlrState*	slrcanonstate(CfgGram<void> *g, SlrState *z);
int		slrcanreduce(CfgGram<void> *g, SlrParse *p, SlrState *z, CfgSym<void> *s, CfgRule<void> **nr);
int		slrcanshift(CfgGram<void> *g, SlrState *z, CfgSym<void> *s, SlrState **nz);
void		slrcheckstate(SlrState *z);
int		slrcompile(CfgGram<void> *g);
int		slrcompilestate(CfgGram<void> *g, SlrState *z);
int		slrcompilestates(CfgGram<void> *g);
SlrState*	slrcomputeshift(CfgGram<void> *g, SlrState *z, CfgSym<void> *s);
void		slrdumptables(CfgGram<void> *g);
void		slrerror(CfgGram<void> *g, SlrParse *p, char *msg);
void		slrfreestate(SlrState *z);
SlrState*	slrinitstate(CfgGram<void> *g, CfgSym<void> *start);
int		slrstatecmp(SlrState *z, SlrState *y);
SlrState*	slrsyncstate(CfgGram<void> *g, SlrState *z);
