#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <stdarg.h>
#include <wchar.h>

int
panic(char *msg)
{
	fprintf(stderr, "panic: %s\n", msg);
	exit(1);
}

int
main()
{
	printf("// Automatically generated by " __FILE__ ", do not edit.\n\n");
	printf(
		"struct Machine\n{\n"
		"\tgcc_version: string;\n"
		"\tgcc_version_major: int;\n"
		"\tgcc_version_minor: int;\n"
		"\tsizeof_char      : int;\n"  // C99 says this should be 1
		"\tsizeof__Bool     : int;\n"
		"\tsizeof_short     : int;\n"
		"\tsizeof_int       : int;\n"
		"\tsizeof_long      : int;\n"
		"\tsizeof_vlong     : int;\n"
		"\tsizeof_float     : int;\n"
		"\tsizeof_double    : int;\n"
		"\tsizeof_ldouble   : int;\n"
		"\tsizeof_ptr       : int;\n"
		"\tsizeof_void      : int;\n"  // gcc allows these, believe it or not
		"\tsizeof_fn        : int;\n"  //
		"\tsizeof_size_t    : int;\n"
		"\tsizeof_ptrdiff_t : int;\n"
		"\tsizeof_wchar_t   : int;\n"
		"\tsizeof_va_list   : int;\n"
		"};\n\n");

	printf("// Do not mutate these variables.\n");
	printf("const machine := Machine(\n");
	printf("\t.gcc_version = \"%s\",\n", __VERSION__);
	printf("\t.gcc_version_major = %d,\n", __GNUC__);
	printf("\t.gcc_version_minor = %d,\n", __GNUC_MINOR__);

	printf("\t.sizeof_char      = %zu,\n", sizeof(char));
	printf("\t.sizeof__Bool     = %zu,\n", sizeof(_Bool));
	printf("\t.sizeof_short     = %zu,\n", sizeof(short));
	printf("\t.sizeof_int       = %zu,\n", sizeof(int));
	printf("\t.sizeof_long      = %zu,\n", sizeof(long));
	printf("\t.sizeof_vlong     = %zu,\n", sizeof(long long));
	printf("\t.sizeof_float     = %zu,\n", sizeof(float));
	printf("\t.sizeof_double    = %zu,\n", sizeof(double));
	printf("\t.sizeof_ldouble   = %zu,\n", sizeof(long double));
	printf("\t.sizeof_ptr       = %zu,\n", sizeof(void *));
	printf("\t.sizeof_void      = %zu,\n", sizeof(void));
	printf("\t.sizeof_fn        = %zu,\n", sizeof(main));
	printf("\t.sizeof_size_t    = %zu,\n", sizeof(size_t));
	printf("\t.sizeof_ptrdiff_t = %zu,\n", sizeof(ptrdiff_t));
	printf("\t.sizeof_wchar_t   = %zu,\n",  sizeof(wchar_t));
	printf("\t.sizeof_va_list   = %zu\n",  sizeof(va_list));
	printf(");\n\n");

	// Try to determine reasonable representations for some
	// built-in types.  Prefer 'long' (see C99 7.17.4).  We
	// should probably pick these off the actual source file,
	// and just use this as a fallback.

	if(sizeof(unsigned long) == sizeof(size_t))
		printf("const Tsize_t    := Tulong;\n");
	else if(sizeof(unsigned long long) == sizeof(size_t))
		printf("const Tsize_t    := Tuvlong;\n");
	else if(sizeof(unsigned int) == sizeof(size_t))
		printf("const Tsize_t    := Tuint;\n");
	else if(sizeof(unsigned short) == sizeof(size_t))
		printf("const Tsize_t    := Tushort;\n");
	else
		panic("weird type for size_t");

	if(sizeof(signed long) == sizeof(ptrdiff_t))
		printf("const Tptrdiff_t := Tlong;\n");
	else if(sizeof(signed long long) == sizeof(ptrdiff_t))
		printf("const Tptrdiff_t := Tvlong;\n");
	else if(sizeof(signed int) == sizeof(ptrdiff_t))
		printf("const Tptrdiff_t := Tint;\n");
	else if(sizeof(signed short) == sizeof(ptrdiff_t))
		printf("const Tptrdiff_t := Tshort;\n");
	else
		panic("weird type for ptrdiff_t");

	// Char types, which C99 says may be signed or unsigned
	printf("const Tchar      := T%cchar;\n", 
	      ((char)0xff) > 0 ? 'u' : 's');

	if(sizeof(signed long long) == sizeof(wchar_t)){
		if(((wchar_t)0xffffffffffffffffULL) > 0)
			printf("const Twchar_t   := Tuvlong;\n");
		else
			printf("const Twchar_t   := Tvlong;\n");
	}
	else if(sizeof(signed int) == sizeof(wchar_t)){
		if(((wchar_t)0xffffffffULL) > 0)
			printf("const Twchar_t   := Tuint;\n");
		else
			printf("const Twchar_t   := Tint;\n");
	}
	else if(sizeof(signed short) == sizeof(wchar_t)){
		if(((wchar_t)0xffffULL) > 0)
			printf("const Twchar_t   := Tushort;\n");
		else
			printf("const Twchar_t   := Tshort;\n");
	}
	else{
		if(((wchar_t)0xffULL) > 0)
			printf("const Twchar_t   := Tuchar;\n");
		else
			printf("const Twchar_t   := Tschar;\n");
	}

	// String types
	printf("const Tstring    := canonptr(Tchar);\n");
	printf("const Twstring_t := canonptr(Twchar_t);\n");

	printf("\n");
	return 0;
} 

