#include "a.h"

DEFINE_bool(profile, false, "Enable profiler.");
DEFINE_bool(profileextra, false, "Enable extra profiling (slow).");

void
profilecall(Fn *caller, Fn *callee)
{
	if(!FLAGS_profile)
		return;
	callee = callee->link;
	assert(callee == callee->link);
	callee->calls++;
	if(FLAGS_profileextra){
		assert(caller == caller->link);
		hashput(caller->callouts, callee, hashget(caller->callouts, callee)+1);
	}
}

void
profileclosure(Fn *creator, Fn *closure)
{
	if(!FLAGS_profile)
		return;
	assert(creator == creator->link);
	creator->closures++;
}

#if 0
static int
byclosures(Fn *a, Fn *b)
{
	return b->closures - a->closures;
}
#endif

static int
bycalls(Fn *a, Fn *b)
{
	return b->calls - a->calls;
}

static int
isntclosure(Fn *fn)
{
	return fn->link == fn;
}

void
printprofile()
{
	List<Fn*> *fns = listfilter(allfns, isntclosure);
	for(List<Fn*> *l = listsort(fns, bycalls); l; l=l->next){
		Fn *fn = l->x;
		print("%7d %7d %s %L\n", fn->calls, fn->closures, fn->name, fn->line);
	}
}

#if 0
static Fn*
runningfn(void)
{
	Thread *t;
	void **fp;
	Fn *fn;

	t = running;
	if(t == nil || t->fp_ == nil)
		return nil;
	fp = (void**)t->fp_;
	fn = (Fn*)fp[FnIndex];
	if(fn == nil)
		return nil;
	while(fn->link)
		fn = fn->link;
	return fn;
}

static int ticks;

void
alrm(int s)
{
	Fn *fn;

	fn = runningfn();
	if(fn)
		hashput(fnticks, fn, hashget(fnticks, fn)+1);
	
	if(ticks++ % 5 == 0)
		tracealrm(s);
}

static List<Fn*>*
sortprofile(Hash<Fn*,uintptr> *h)
{
	Hashkv<Fn*,uintptr> kv;
	Hashiter<Fn*,uintptr> iter;
	Fn *fn;
	uint val;
	List<Fn*> *sorted, *l, *ll;

	if(!h)
		return nil;
	sorted = ZList;
	hashiterstart(&iter, h);
	while(hashiternextkv(&iter, &kv)){
		fn = kv.key;
		val = kv.value;
		for(ll=nil, l=sorted; l && val < hashget(h, l->x); ll=l, l=l->next)
			;
		if(ll)
			ll->next = listinsert(nil, ll->next, fn);
		else
			sorted = listinsert(nil, sorted, fn);
	}
	return sorted;
}

int
digits(int n)
{
	int d;
	
	d = 1;
	while(n >= 10){
		n /= 10;
		d++;
	}
	return d;
}

void dodot(void);

void
printprofile(void)
{
	List<Fn*> *sorted, *l;
	Fn *fn;
	int n, t, c;
	char where[100];
	int callwid, tickwid;

	fprint(2, "Profile [%d ticks]:\n", ticks);
	if(ticks == 0)
		ticks = 1;
	callwid = digits(calls);
	if(callwid < 5)
		callwid = 5;
	tickwid = digits(ticks);
	fprint(2, "%*s  %*s  %-30s  %s\n",
		5+1+tickwid, "Time",
		callwid, "Calls",
		"Source",
		"Function");
	sorted = sortprofile(fnticks);
	n = 50;
	for(l=sorted; l && n-- > 0; l=l->next){
		fn = l->x;
		t = hashget(fnticks, fn);
		c = hashget(fncalls, fn);
		if(fn->builtin)
			strcpy(where, "[builtin]");
		else
			snprint(where, sizeof where, "%L", fn->line);
		fprint(2, "%4.1f%% %*d  %*d  %-30s  %s\n",
			100.0*t/ticks, tickwid, t, callwid, c, where, fn->name);
	}
	
	dodot();
}
// Clumsy stack trace profiler.  Runs at 1/5 frequency of
// regular profiler because it does so much more work
// and is not particularly efficient.  Ideas from google-perftools
// but the code is new (and pretty bad).

enum
{
	MaxTraceFn = 32,
	TraceCount = 4,
	TraceHashSize = 65536,
};

typedef struct Trace Trace;
struct Trace
{
	Fn *fn[MaxTraceFn];
	int n;
	int count;
	int age;
};

static int
getstacktrace(Fn **fnp, int n)
{
	int i;
	Thread *t;
	Fn *fn;
	void **fp;
	
	t = running;
	if(t == nil || t->fp_ == nil)
		return -1;
	fp = (void**)t->fp_;
	for(i=0; i<n && fp && (fn = (Fn*)fp[FnIndex]); i++){
		fnp[i] = fn;
		fp = (void**)fp[FpIndex];
	}
	if(i < n)
		fnp[i] = nil;
	return i;
}

static Trace **tracehash;
static List<Trace*> *overflows;
static Trace *tbuf, *tbuf0;
static int ntbuf;
static int age;

static uint64
llhash(uint64 key)
{
	/* derived from http://www.concentric.net/~Ttwang/tech/inthash.htm */
	key += ~(key << 32);
	key ^= (key >> 22);
	key += ~(key << 13);
	key ^= (key >> 8);
	key += (key << 3);
	key ^= (key >> 15);
	key += ~(key << 27);
	key ^= (key >> 31);
	return key;
}
static uint32
hash2(uint32 h, uint32 i)
{
	return llhash(((uint64)h<<32)|i);
}

static uint
hashtrace(Fn **fn, int n)
{
	int i;
	uint32 h;
	
	h = 0x12345678;
	for(i=0; i<n; i++)
		h = hash2(h, (uint32)(uintptr)fn[i]);
	return h;
}

void
tracealrm(int s)
{
	Fn *fn[MaxTraceFn];
	int i, n, young;
	uint h;
	Trace *t;
	
	if(tracehash == nil)
		tracehash = (Trace**)emalloc(TraceHashSize * sizeof tracehash[0]);
	n = getstacktrace(fn, nelem(fn));
	if(n <= 0)
		return;
	h = hashtrace(fn, n) % TraceHashSize;
	
	t = tracehash[h];
	if(t == nil){
		t = (Trace*)emalloc(TraceCount * sizeof *t);
		tracehash[h] = t;
	}
	
	young = -1;
	for(i=0; i<TraceCount; i++){
		if(t[i].n == n && memcmp(t[i].fn, fn, n*sizeof fn[0]) == 0){
			t[i].count++;
			t[i].age = age++;
			return;
		}
		if(young == -1 || t[i].age < t[young].age || t[i].n == 0)
			young = i;
	}
	
	t += young;
	if(t->n > 0){
		if(ntbuf == 0){
			if(tbuf)
				overflows = listinsert(nil, overflows, tbuf0);
			tbuf0 = (Trace*)emalloc(1000*sizeof tbuf[0]);
			ntbuf = 999;
			tbuf = tbuf0;
		}
		*tbuf = *t;
		tbuf++;
		ntbuf--;
	}

	t->n = n;
	memmove(t->fn, fn, n*sizeof fn[0]);
	t->age = age++;
	t->count = 1;
}

Hash<Fn*,Fn*> *fns;

void
addtracecount(Trace *t)
{
	int i;
	Fn *fn, *down;
	
	if(fns == nil)
		fns = mkhash<Fn*,Fn*>();

	down = nil;
	for(i=0; i<t->n; i++){
		fn = t->fn[i];
		hashput(fns, fn, fn);
		fn->cticks += t->count;
		if(fn->callhash == nil)
			fn->callhash = mkhash<Fn*,uintptr>();
		if(down)
			hashput(fn->callhash, down, hashget(fn->callhash, down) + t->count);
		else
			fn->ticks += t->count;
		down = fn;
	}
}

void
dotprofile(int fd)
{
	int i, j, count;
	List<Trace*> *l;
	Trace *t;
	Fn *fn, *down;
	Fmt fmt;
	char buf[8192];
	double frac;
	int tticks;
	
	if(ticks == 0)
		return;

	tticks = (ticks+4)/5;
	if(tticks < 1)
		tticks = 1;

	for(l=overflows; l; l=l->next){
		t = l->x;
		for(; t->n > 0; t++)
			addtracecount(t);
	}
	for(t=tbuf0; t<tbuf; t++)
		addtracecount(t);
	for(i=0; i<TraceHashSize; i++){
		t = tracehash[i];
		if(t == nil)
			continue;
		for(j=0; j<TraceCount; j++)
			if(t[j].n)
				addtracecount(&t[j]);
	}

	fmtfdinit(&fmt, fd, buf, sizeof buf);
	fmtprint(&fmt, "digraph \"%d samples\" {\n", tticks);
	fmtprint(&fmt, "node [width=0.375,height=0.25]\n");
	// XXX legend here
	Hashiter<Fn*,Fn*> iter;
	Hashkv<Fn*,Fn*> kv;
	hashiterstart(&iter, fns);
	while(hashiternextkv(&iter, &kv)){
		fn = kv.key;
		if(fn->cticks < 5)
			continue;
		frac = 8;
		if(fn->ticks > 0)
			frac = 8 + (50.0 * sqrt(fabs(fn->ticks * 1.0 / tticks)));

		fmtprint(&fmt, "N%p [label=\"%s\\n%d (%3.1f%%)\\rof %d (%3.1f%%)\\r\",shape=box,fontsize=%.1f]\n",
			fn, fn->name, fn->ticks, 100.0 * fn->ticks / tticks, 
			fn->cticks, 100.0 * fn->cticks / tticks,
			frac);
	}
	
	hashiterstart(&iter, fns);
	while(hashiternextkv(&iter, &kv)){
		fn = (Fn*)kv.key;
		Hashiter<Fn*,uintptr> iter1;
		Hashkv<Fn*,uintptr> kv1;
		hashiterstart(&iter1, fn->callhash);
		while(hashiternextkv(&iter1, &kv1)){
			down = kv1.key;
			count = kv1.value;
			if(count < 5)
				continue;
			frac = fabs(3. * count / tticks);
			if(frac > 1)
				frac = 1;
			frac *= 2;
			if(down)
				fmtprint(&fmt, "N%p -> N%p [label=%d, weight=%d, style=\"setlinewidth(%f)\"];\n",
					fn, down, count, (int)pow(count, 0.7), frac);
		}
	}
	fmtprint(&fmt, "}\n");
	fmtfdflush(&fmt);
}

void
dodot(void)
{
	int fd;
	
	if((fd = creat("/tmp/zeta-profile.dot", 0666)) < 0)
		panic("creat /tmp/zeta-profile.dot");
	dotprofile(fd);
	close(fd);
	
	system("dot -Tps /tmp/zeta-profile.dot >/tmp/zeta-profile.ps && gv /tmp/zeta-profile.ps");
}



#endif
